/* Basic styling */
/* Draw the lines */
.jOrgChart .line {
  height                : 20px;
  width                 : 2px;
}

.jOrgChart .down {
  background-color      : #999;
  margin                : 0px auto;
}

.jOrgChart .top {
  border-top            : 2px solid #999;
}

.jOrgChart .left {
  border-right          : 2px solid #999;
}

.jOrgChart .right {
  border-left           : 2px solid #999;
}

/* node cell */
.jOrgChart td {
  text-align            : center;
  vertical-align        : top;
  padding               : 0;
}

/* The node */
.jOrgChart .node {
  background-color      : #FFF;
  border                : 3px solid #999;
  display               : inline-block;
  width                 : 250px;
  height                : 100px;
  z-index               : 10;
  margin                : 0 2px;
  margin-top            : 0 4px;
  border-radius         : 10px;
}

.jOrgChart .node.main {
  background-color      : #F5F5F5;
}

.jOrgChart a {
  color                 : black;
}

.jOrgChart a.title {
  font-weight           : bold;
}

.jOrgChart p.dates {
  font-size             : 80%;
}

div.other-parents {
  margin-top: 20px;
}
